﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.uicommand;
/*
 * UIコマンド
 * 
 * C.xxxxxxxx
 */

private import pes.byteorder;
private import std.string;
private import std.path;
private import common.nodetable;

// コマンドコード
enum UICommandCode : uint
{
	QSTART_CORE			= 0x00000000,		// コア起動
	RSTART_CORE			= 0x10000000,		// コア起動(応答)
	QSET_LISTENER_PORT	= 0x00000001,		// リスナーポート設定
	RSET_LISTENER_PORT	= 0x10000001,		// リスナーポート設定(応答)
	QGET_LISTENER_PORT	= 0x00000002,		// リスナーポート取得
	RGET_LISTENER_PORT	= 0x10000002,		// リスナーポート取得（応答）
	QSET_SPEED			= 0x00000003,		// 回線速度設定
	RSET_SPEED			= 0x10000003,		// 回線速度設定（応答）
	QGET_SPEED			= 0x00000004,		// 回線速度取得
	RGET_SPEED			= 0x10000004,		// 回線速度取得
	QGET_CLUSTERING_KEY	= 0x00000005,		// クラスタリングワード取得
	RGET_CLUSTERING_KEY	= 0x10000005,		// クラスタリングワード取得(応答)
	QSET_CLUSTERING_KEY	= 0x00000006,		// クラスタリングワード設定
	RSET_CLUSTERING_KEY	= 0x10000006,		// クラスタリングワード設定(応答)
	QGET_NODE			= 0x00000010,		// ノード取得
	RGET_NODE			= 0x10000010,		// ノード取得
	QADD_SEARCH_WORD	= 0x00000020,		// 検索ワード登録
	RADD_SEARCH_WORD	= 0x10000020,		// 検索ワード登録(応答)
	QREMOVE_SEARCH_WORD	= 0x00000021,		// 検索ワード削除
	RREMOVE_SEARCH_WORD	= 0x10000021,		// 検索ワード削除(応答)
	QSEARCH				= 0x00000024,		// 検索
	RSEARCH				= 0x10000024,		// 検索(応答)
}

enum UIStatus: uint
{
	UNDEFINED		= 0,		// 未定義
	OK				= 200,		// 成功
	NG				= 300,		// 失敗
	COMMAND_ERROR	= 400,		// コマンドエラー
								// (コマンドコードエラー、フォーマットエラーなど)
}

// 可変文字列
/*
struct VarString{
	uint length;
	ubyte[length] data;
}
*/

// 可変長文字列 -> バイト列
ubyte[] toVarString(char[] data)
{
	uint length = hostToNetwork32(data.length);
	return (cast(ubyte *)&length)[0 .. uint.sizeof] ~ cast(ubyte[])data;
}

// バイト列 -> 可変長文字列
char[] fetchVarString(inout ubyte[] data)
{
	char[] str;
	uint str_length = networkToHost32(*cast(uint *)data.ptr);
	
	if (str_length + uint.sizeof > data.length) {
		throw new Exception("バイト長が長すぎ");
	}
	str		= cast(char[])data[uint.sizeof .. uint.sizeof + str_length].dup;
	data	= data[uint.sizeof + str_length .. length];
	
	return str;
}


// リクエストベースクラス(UI->CORE)
class UIRequestCommand
{
protected:
	uint		m_command_code;
	uint		m_length;
	ubyte[]		m_data;
	
public:
	this() {}
	this(ubyte[] command_block)
	{
		unpack(command_block);
	}
	
	void unpack(ubyte[] command_block)
	{
		if (command_block.length < uint.sizeof * 2) {
			throw new Exception("データサイズ小さすぎ");
		}
		m_data			= null;
		m_command_code 	= networkToHost32(*cast(uint*)command_block.ptr);
		m_length		= networkToHost32(*cast(uint*)(command_block.ptr + uint.sizeof));
		if (command_block.length > uint.sizeof * 2) {
			m_data = command_block[uint.sizeof * 2 .. length].dup;
		}
		if (m_length != m_data.length) {
			throw new Exception("データサイズが不正");
		}
	}
	
	void assign(UIRequestCommand command)
	{
		m_command_code 	= command.commandCode;
		m_length		= command.length;
		m_data			= command.data;
	}
	
	UICommandCode commandCode()
	{
		return cast(UICommandCode)m_command_code;
	}
	
	uint length()
	{
		return m_length;
	}
	
	ubyte[] data()
	{
		return m_data;
	}
}

// レスポンスベースクラス (CORE->UI)
class UIResponseCommand
{
protected:
	uint	m_command_code;
	uint	m_status;
	uint	m_length;
	ubyte[] m_data;
	
public:
	this() {}
	ubyte[] pack()
	{
		ubyte[] data;
		uint	command_code 	= hostToNetwork32(m_command_code);
		uint	length			= hostToNetwork32(m_data.length);
		uint	status			= hostToNetwork32(m_status);
		
		data = (cast(ubyte*)&command_code)[0 .. uint.sizeof].dup
				~ (cast(ubyte*)&status)[0 .. uint.sizeof].dup
				~ (cast(ubyte*)&length)[0 .. uint.sizeof].dup
				~ m_data;
		
		//delete m_data, m_data = null;
		m_data = null;
		
		return data;
	}
	
	void commandCode(UICommandCode code)
	{
		m_command_code = code;
	}
	
	void status(UIStatus status)
	{
		m_status = cast(uint)status;
	}
	
	void data(ubyte[] data)
	{
		m_data = data;	// do not dup
	}
}

// エラー応答
class UIRError: UIResponseCommand
{
protected:
	char[]	m_error_message;
	
public:
	void errorMessage(char[] message)
	{
		m_error_message = message.dup;
	}
	
	override ubyte[] pack()
	{
		data(toVarString(m_error_message));
		return super.pack();
	}
}

// リスナーポート設定
class UIQSetListenerPort: UIRequestCommand
{
protected:
	ushort	m_port;
	
	void unpack()
	{
		ubyte[] data = super.data;
		if (length < uint.sizeof) {
			throw new Exception("データサイズ小さすぎ");
		}
		uint port = networkToHost32(*cast(uint *)data.ptr);
		
		m_port = cast(ushort)port;
	}
	
public:
	this(UIRequestCommand command)
	{
		assign(command);
		unpack();
	}
	
	ushort port()
	{
		return m_port;
	}
	
}

// リスナーポート設定応答
class UIRSetListenerPort: UIResponseCommand
{
	this()
	{
		commandCode = UICommandCode.RSET_LISTENER_PORT;
	}
}


// リスナーポート取得
class UIQGetListenerPort: UIRequestCommand
{
protected:
public:
	this(UIRequestCommand command)
	{
		assign(command);
	}
}


// リスナーポート取得応答
class UIRGetListenerPort: UIResponseCommand
{
protected:
	uint m_port;

public:
	this()
	{
		commandCode = UICommandCode.RGET_LISTENER_PORT;
	}
	
	override ubyte[] pack()
	{
		uint port = hostToNetwork32(m_port);
		data((cast(ubyte*)&port)[0 .. port.sizeof]);
		
		return super.pack();
	}
	
	void port(ushort port)
	{
		m_port = cast(uint)port;
	}
}

// 回線速度設定
class UIQSetSpeed: UIRequestCommand
{
protected:
	ushort	m_speed;
	
	void unpack()
	{
		ubyte[] data = super.data;
		if (length < uint.sizeof) {
			throw new Exception("データサイズ小さすぎ");
		}
		uint speed = networkToHost32(*cast(uint *)data.ptr);
		
		m_speed = cast(ushort)speed;
	}
	
public:
	this(UIRequestCommand command)
	{
		assign(command);
		unpack();
	}
	
	ushort speed()
	{
		return m_speed;
	}
	
}

// 回線速度設定応答
class UIRSetSpeed: UIResponseCommand
{
	this()
	{
		commandCode = UICommandCode.RSET_SPEED;
	}
}

// 回線速度取得
class UIQGetSpeed: UIRequestCommand
{
protected:
public:
	this(UIRequestCommand command)
	{
		assign(command);
	}
}

// 回線速度取得応答
class UIRGetSpeed: UIResponseCommand
{
protected:
	uint m_speed;

public:
	this()
	{
		commandCode = UICommandCode.RGET_SPEED;
	}
	
	override ubyte[] pack()
	{
		uint speed = hostToNetwork32(m_speed);
		data((cast(ubyte*)&speed)[0 .. speed.sizeof]);
		
		return super.pack();
	}
	
	void speed(uint speed)
	{
		m_speed = cast(uint)speed;
	}
}


// クラスタリングワード設定
class UIQSetClusteringKey: UIRequestCommand
{
protected:
	char[][3] m_clustering_key;
	
	void unpack()
	{
		ubyte[] data = super.data;
		
		m_clustering_key[0] = fetchVarString(data);
		m_clustering_key[1] = fetchVarString(data);
		m_clustering_key[2] = fetchVarString(data);
	}
	
public:
	this(UIRequestCommand command)
	{
		assign(command);
		unpack();
	}
	
	char[] clusteringKey(uint i)
	in {
		assert(0 <= i && i < 3);
	}
	body {
		return m_clustering_key[i];
	}
	
	char[][] clusteringKey()
	{
		return m_clustering_key.dup;
	}
}

// クラスタリングワード設定(応答)
class UIRSetClusteringKey: UIResponseCommand
{
	this()
	{
		commandCode = UICommandCode.RSET_CLUSTERING_KEY;
	}
}

// クラスタリングワード取得
class UIQGetClusteringKey: UIRequestCommand
{
protected:
public:
	this(UIRequestCommand command)
	{
		assign(command);
	}
}

// クラスタリングワード取得(応答)
class UIRGetClusteringKey: UIResponseCommand
{
protected:
	char[][3]	m_clustering_key;

public:
	this()
	{
		commandCode = UICommandCode.RGET_CLUSTERING_KEY;
	}
	
	override ubyte[] pack()
	{
		m_data  = toVarString(m_clustering_key[0]);
		m_data ~= toVarString(m_clustering_key[1]);
		m_data ~= toVarString(m_clustering_key[2]);
		
		return super.pack();
	}
	
	void clusteringKey(uint i, char[] key)
	in {
		assert(0 <= i && i < 3);
	}
	body {
		m_clustering_key[i] = key;
	}
	
	void clusteringKey(char[][] keys)
	in {
		assert(keys.length == m_clustering_key.length);
	}
	body {
		for (int i = 0; i < m_clustering_key.length; ++i) {
			m_clustering_key[i] = keys[i];
		}
	}
}


// 検索登録
class UIQAddSearchWord: UIRequestCommand
{
protected:
	char[]	m_word;
	
	void unpack()
	{
		ubyte[] data 	= super.data;
		m_word			= fetchVarString(data);
	}
	
public:
	this(UIRequestCommand command)
	{
		assign(command);
		unpack();
	}
	
	char[] word()
	{
		return m_word;
	}
}

// 検索登録応答
class UIRAddSearchWord: UIResponseCommand
{
public:
	this()
	{
		commandCode = UICommandCode.RGET_SPEED;
	}
}

// 検索削除
class UIQRemoveSearchWord: UIRequestCommand
{
protected:
	char[]	m_word;
	
	void unpack()
	{
		ubyte[] data	= super.data;
		m_word			= fetchVarString(data);
	}
	
public:
	this(UIRequestCommand command)
	{
		assign(command);
		unpack();
	}
	
	char[] word()
	{
		return m_word;
	}
}

// 検索削除応答
class UIRRemoveSearchWord: UIResponseCommand
{
public:
	this()
	{
		commandCode = UICommandCode.RREMOVE_SEARCH_WORD;
	}
}


enum UISearchKeyType: uint
{
	HASH		= 0,	// ファイル全体ハッシュ
	KEYWORD		= 1,	// ファイル名マッチ
	TRIP		= 2,	// トリップ
}

enum UIFileSortType: uint
{
	FILE_NAME	= 0,	// ファイル名
	FILE_SIZE	= 1,	// ファイルサイズ
	REFF_COUNT	= 2,	// 非参照割合(人気度)
	FILE_EXT	= 3, 	// 拡張子
	TAG			= 4, 	// タグ
}


class UIQSearch: UIRequestCommand
{
protected:
	uint				m_max;
	UIFileSortType		m_sort_type;
	UISearchKeyType		m_key_type;
	char[]				m_key;
	
	void unpack()
	{
		ubyte[] data	= super.data;
		if (data.length <= m_max.sizeof + m_sort_type.sizeof + m_key_type.sizeof + uint.sizeof) {
			throw new Exception("コマンドが不正");
		}
		m_max			= networkToHost32(*cast(uint*)data.ptr);
		m_sort_type		= cast(UIFileSortType)networkToHost32(*cast(uint*)(data.ptr + uint.sizeof));
		m_key_type 		= cast(UISearchKeyType)networkToHost32(*cast(uint*)(data.ptr + uint.sizeof * 2));
		data			= data[uint.sizeof * 3 .. length];
		m_key			= fetchVarString(data);
	}
	
public:
	this(UIRequestCommand command)
	{
		assign(command);
		unpack();
	}
	this(){}
	
	char[] key()
	{
		return m_key;
	}
	
	UISearchKeyType keyType()
	{
		return m_key_type;
	}
	
	UIFileSortType sortType()
	{
		return m_sort_type;
	}
	
	uint max()
	{
		return m_max;
	}
	
	// setter
	void key(char[] key)
	{
		m_key = key.dup;
	}
	
	void keyType(UISearchKeyType type)
	{
		m_key_type = type;
	}
	
	void sortType(UIFileSortType type)
	{
		m_sort_type = type;
	}
	
	void max(uint max)
	{
		m_max = max;
	}
}

class UIFileInformation 
{
protected:
	uint			m_file_size;		// ファイルサイズ
	uint			m_referred_count;	// 非参照数
	char[10]		m_trip;				// トリップ(アルファベットor数字以外が含まれるまでがトリップ)
	ubyte[16]		m_hash;				// ファイル全体のMD5 128bitハッシュ値
	char[]			m_filename;			// ファイル名(UTF-8)
	Object			m_tag;				// tag
	
	UIFileSortType	m_sort_type;
	bool			m_asc;
	
public:
	
	void asc()
	{
		m_asc = true;
	}
	
	void desc()
	{
		m_asc = false;
	}
	
	ubyte[] pack()
	{
		ubyte[] data;
		uint	netbyte;
		
		netbyte		= hostToNetwork32(m_file_size);
		data 		~= (cast(ubyte *)&netbyte)[0 .. netbyte.sizeof].dup;
		netbyte		= hostToNetwork32(m_referred_count);
		data 		~= (cast(ubyte *)&netbyte)[0 .. netbyte.sizeof].dup;
		data		~= cast(ubyte[])m_trip;
		data		~= m_hash;
		data		~= toVarString(m_filename);
		
		return data;
	}
	
	int opCmp(Object rhs)
	{
		UIFileInformation rhs_file = cast(UIFileInformation)rhs;
		if (rhs_file is null || rhs_file is this) {
			return 0;
		}
		if (m_asc) {
			switch (m_sort_type) {
			case UIFileSortType.FILE_NAME:
				return std.string.cmp(fileName, rhs_file.fileName);
				
			case UIFileSortType.FILE_SIZE:
				return (cast(long)fileSize) - (cast(long)rhs_file.fileSize);
				
			case UIFileSortType.REFF_COUNT:
				return popularity - rhs_file.popularity;
			case UIFileSortType.FILE_EXT:
				return std.string.cmp(fileExt, rhs_file.fileExt);
			case UIFileSortType.TAG:
				return tag.opCmp(rhs_file.tag);
			}
		} else {
			switch (m_sort_type) {
			case UIFileSortType.FILE_NAME:
				return std.string.cmp(rhs_file.fileName, fileName);
				
			case UIFileSortType.FILE_SIZE:
				return (cast(long)rhs_file.fileSize) - (cast(long)fileSize);
			case UIFileSortType.REFF_COUNT:
				return rhs_file.popularity - popularity;
			case UIFileSortType.FILE_EXT:
				return std.string.cmp(rhs_file.fileExt, fileExt);
			case UIFileSortType.TAG:
				return rhs_file.tag.opCmp(tag);
			}
		}
		
		return 0;
	}
	
	// set
	
	Object tag()
	{
		return m_tag;
	}
	
	void tag(Object o)
	{
		m_tag = o;
	}
	
	void fileSize(uint size)
	{
		m_file_size = size;
	}
	
	uint fileSize()
	{
		return m_file_size;
	}
	
	void referredCount(uint refcnt)
	{
		m_referred_count = refcnt;
	}
	
	int popularity()
	{
		return (referredCount * 0x10000) / ((fileSize < 0x10000) ? 0x10000:fileSize);
	}
	
	uint referredCount()
	{
		return m_referred_count;
	}
	
	void trip(char[] trip)
	{
		m_trip[0 .. length] = 0;
		if (trip.length <= m_trip.length) {
			m_trip[0 .. trip.length] = trip[0 .. length];
		}
	}
	
	char[] trip()
	{
		return m_trip;
	}
	
	ubyte[] hash()
	{
		return m_hash;
	}
	
	void hash(ubyte[] hash)
	in {
		assert(hash.length == m_hash.length);
	}
	body {
		m_hash[0 .. length] = hash[0 .. m_hash.length];
	}
	
	void fileName(char[] name)
	{
		m_filename = name;
	}
	
	char[] fileName()
	{
		return m_filename;
	}
	
	char[] fileExt()
	{
		return std.path.getExt(m_filename);
	}
	
	void sortType(UIFileSortType type)
	{
		m_sort_type = type;
	}
	
	static void setSortType(UIFileInformation[] array, UIFileSortType type)
	{
		foreach (UIFileInformation e; array) {
			e.sortType = type;
		}
	}
	
	static void setSortAscent(UIFileInformation[] array, bool asc)
	{
		foreach (UIFileInformation e; array) {
			if (asc) {
				e.asc();
			} else {
				e.desc();
			}
		}
	}
	
	void destroy()
	{
		delete m_filename;
		delete m_tag;
		m_filename = null;
		m_tag = null;
	}
}


// 検索結果応答
class UIRSearch: UIResponseCommand
{
protected:
	uint				m_max;
	UIFileInformation[] m_record;
	
public:
	
	this()
	{
		commandCode = UICommandCode.RSEARCH;
	}
	
	override ubyte[] pack()
	{
		uint i;
		uint field_num = hostToNetwork32(m_max < m_record.length ? m_max: m_record.length);
		m_data 	= null;
		
		m_data ~= (cast(ubyte *)&field_num)[0 .. field_num.sizeof].dup;
		
		
		field_num = networkToHost32(field_num);
		foreach (UIFileInformation e; m_record)
		{
			m_data ~= e.pack();
			
			if (i++ >= field_num) {
				break;
			}
		}
		
		return super.pack();
	}
	
	void record(UIFileInformation[] info)
	{
		m_record = info;	// do not dup
	}
	
	UIFileInformation[] record()
	{
		return m_record;
	}
	
	void sort(UIFileSortType type)
	{
		UIFileInformation.setSortType(m_record, type);
		m_record = m_record.sort;
	}
	
	void max(uint max)
	{
		m_max = max;
	}
}

enum UINodeType: uint {
	UNCONNECTION			= 1,
	DATA_CONNECTION			= 2,
	DOWNSTREAM_CONNECTION	= 4,
	UPSTREAM_CONNECTION		= 8,
}

class UINodeDetails
{
	UINodeType	nodeType	= UINodeType.UNCONNECTION;
	NodeDetails		basic;
	uint				rate;
	uint				ctl;
	
	int opCmp(Object rhs)
	{
		UINodeDetails rhsnode = cast(UINodeDetails)rhs;
		
		if (rhsnode.nodeType > nodeType) {
			return 1;
		} else if (rhsnode.nodeType < nodeType) {
			return -1;
		}
		
		return rhsnode.basic.sortKey - basic.sortKey;
	}
	
	this(UINodeType type, NodeDetails element, uint ctl = 0, uint rate = 0)
	{
		nodeType	= type;
		basic		= element;
		this.ctl	= ctl;
		this.rate	= rate;
	}
}

// ノード情報取得要求
class UIQGetNodeInformation: UIRequestCommand
{
protected:
	uint		m_node_type;
	
	void unpack()
	{
		ubyte[] data		= super.data;
		
		if (data.length > uint.sizeof) {
			throw new Exception("コマンドが不正");
		}
		
		m_node_type = networkToHost32(*cast(uint*)super.data.ptr);
		if (! (0 < m_node_type && m_node_type < 16)) {
			throw new Exception("コマンドが不正");
		}
	}
	
public:
	this(UIRequestCommand command)
	{
		assign(command);
		unpack();
	}
	
	uint nodeType()
	{
		return m_node_type;
	}
}

// ノード情報取得応答
class UIRGetNodeInformation: UIResponseCommand
{
protected:
	UINodeDetails[]		m_node_info;
	
	struct NodeRecordH {
	align(1):
		uint	nodeType;
		uint	priority;
		uint	ctl;
		uint	rate;
		uint	address;
		uint	port;
		uint	speed;
		// クラスタリング文字列
	};
	
public:
	this()
	{
		commandCode = UICommandCode.RGET_NODE;
	}
	
	ubyte[] pack()
	{
		NodeRecordH		header;
		uint			count = hostToNetwork32(m_node_info.length);
		
		m_data	= null;
		m_data ~= 		(cast(ubyte *)&count)[0 .. uint.sizeof].dup;
		
		foreach (UINodeDetails e; m_node_info.sort) {
			header.nodeType = hostToNetwork32(e.nodeType);
			header.priority	= hostToNetwork32(e.basic.priority);
			header.ctl		= hostToNetwork32(e.ctl);
			header.rate		= hostToNetwork32(e.rate);
			header.address	= hostToNetwork32(e.basic.inetAddress.addr);
			header.port		= hostToNetwork32(e.basic.inetAddress.port);
			header.speed	= hostToNetwork32(e.basic.speed);
			
			m_data ~= (cast(ubyte *)&header)[0 .. header.sizeof];
			foreach (char[] s; e.basic.clusteringKey) {
				m_data ~= toVarString(s);
			}
		}
		
		return super.pack;
	}
	
	void nodeDetails(UINodeDetails[] info)
	{
		m_node_info = info;
	}
}

